"""
Representação dos dados ansible de uma máquina virtual
"""


class VMAnsible:
    def __init__(self, grupo):
        self.grupo = grupo
        self.variaveis = []

    def extrair_dados_vars_dict(self, dict_vars, nome_vm):
        lista_nomes_var = []
        for item in dict_vars or {}:
            nome_var = item.get('nome')

            if nome_var in lista_nomes_var:
                raise ValueError(
                    (f"Variável '{nome_var}' do grupo ansible '{self.grupo}' "
                     f"referenciada mais de uma vez na VM '{nome_vm}'."))

            lista_nomes_var.append(nome_var)
            ansible_var = VMAnsibleVars(
                nome_var, item.get('valor'))

            self.variaveis.append(ansible_var)

    def __hash__(self):
        return hash(self.grupo)

    def __eq__(self, other):
        return isinstance(other, VMAnsible) and (self.grupo == other.grupo
                                                 and self.variaveis == other.variaveis)

    def __repr__(self):
        return f'''
                grupo: {self.grupo}
                vars: {self.variaveis}
                '''

    def to_dict(self):
        return {
            'grupo': self.grupo,
            'vars': [variavel.to_dict() for variavel in self.variaveis]
        }


class VMAnsibleVars:
    def __init__(self, nome, valor):
        self.nome = nome
        self.valor = valor

    def __eq__(self, other):
        return isinstance(other, VMAnsibleVars) and (self.nome == other.nome
                                                     and self.valor == other.valor)

    def __repr__(self):
        return f'''
                nome: {self.nome}
                valor: {self.valor}
                '''

    def to_dict(self):
        return {
            'nome': self.nome,
            'valor': self.valor
        }



"""
Módulo com funções que tratam da impressão de informações
"""
import sys
from datetime import datetime
from pytz import timezone

# Parâmetros globais de escrita
_EXIBIR_CORES = True


def set_parametros_globais_escrita(exibir_cores):
    global _EXIBIR_CORES  # pylint: disable=global-statement
    _EXIBIR_CORES = exibir_cores


def formatar_msg_aviso(msg):
    msg_inicial = 'AVISO'
    if _EXIBIR_CORES:
        return f'\033[93m{msg_inicial}: {msg}\033[0m'
    return f'{msg_inicial}: {msg}'


def formatar_msg_erro(msg):
    if _EXIBIR_CORES:
        return f'\033[91m{msg}\033[0m'
    return msg


def finalizar_com_erro(msg_erro):
    print(formatar_msg_erro(f'\nErro ao executar operação:\n{msg_erro}'))
    sys.exit(1)


def imprimir_ok(ocultar_progresso):
    if not ocultar_progresso:
        print('[OK]')


def imprimir_erro(ocultar_progresso):
    if not ocultar_progresso:
        print('[ERRO]')


def imprimir_acao_corrente(acao, ocultar_progresso):
    if not ocultar_progresso:
        print(f'{acao:<57}...', end='', flush=True)


def get_str_data_formatada(formato):
    fuso = timezone('America/Sao_Paulo')
    return datetime.now().astimezone(fuso).strftime(formato)



"""
Módulo que realiza o parser de um inventário remoto (no SCVMM)
"""
import json
from vmm_manager.infra.comando import Comando
from vmm_manager.util.config import (CAMPO_AGRUPAMENTO, CAMPO_ID, CAMPO_IMAGEM,
                                     CAMPO_REGIAO, CAMPO_REDE_PRINCIPAL)
from vmm_manager.entidade.inventario import Inventario
from vmm_manager.entidade.vm import VM
from vmm_manager.entidade.vm_rede import VMRede
from vmm_manager.scvmm.enums import VMStatusEnum
from vmm_manager.entidade.vm_disco import VMDisco
from vmm_manager.scvmm.enums import SCDiskBusType, SCDiskSizeType
from vmm_manager.scvmm.scregion import SCRegion


# pylint: disable=too-few-public-methods
class ParserRemoto:

    @staticmethod
    def __get_regioes_disponiveis(servidor_acesso):
        cmd = Comando('obter_regioes_disponiveis',
                      servidor_vmm=servidor_acesso.servidor_vmm)
        status, regioes = cmd.executar(servidor_acesso)
        if not status:
            raise Exception(
                f'Erro ao recuperar regiões disponíveis: {regioes}')

        regioes_remoto = json.loads(regioes)
        regioes_disponiveis = []
        for regiao in regioes_remoto:
            regiao_obj = SCRegion(
                regiao.get('IDNo'),
                regiao.get('NomeNo'),
                regiao.get('Grupo'),
                regiao.get('Cluster')
            )

            regioes_disponiveis.append(regiao_obj)

        return regioes_disponiveis

    def __init__(self, agrupamento, nuvem):
        self.agrupamento = agrupamento
        self.nuvem = nuvem
        self.__inventario = None

    def __get_vms_servidor(self, servidor_acesso, filtro_nome_vm=None):
        cmd = Comando('obter_vms_agrupamento',
                      servidor_vmm=servidor_acesso.servidor_vmm,
                      campo_agrupamento=CAMPO_AGRUPAMENTO[0],
                      campo_id=CAMPO_ID[0],
                      campo_imagem=CAMPO_IMAGEM[0],
                      campo_regiao=CAMPO_REGIAO[0],
                      campo_rede_principal=CAMPO_REDE_PRINCIPAL[0],
                      agrupamento=self.agrupamento,
                      filtro_nome_vm=filtro_nome_vm,
                      nuvem=self.nuvem)
        status, vms = cmd.executar(servidor_acesso)
        if not status:
            raise Exception(
                f"Erro ao recuperar VM's do agrupamento: {vms}")
        return vms

    def __get_discos_adicionais(self, servidor_acesso):
        cmd = Comando('obter_discos_adicionais',
                      servidor_vmm=servidor_acesso.servidor_vmm,
                      campo_agrupamento=CAMPO_AGRUPAMENTO[0],
                      campo_id=CAMPO_ID[0],
                      agrupamento=self.agrupamento,
                      nuvem=self.nuvem,
                      vm_nomes=','.join([f'"{nome_vm}"' for nome_vm in self.__inventario.vms]))
        status, discos_adicionais = cmd.executar(servidor_acesso)
        if not status:
            raise Exception(
                f'Erro ao recuperar discos adicionais: {discos_adicionais}')

        discos_vms_remoto = json.loads(discos_adicionais)
        discos_vms = {}
        for maquina_virtual in discos_vms_remoto:
            nome_vm = maquina_virtual.get('Nome')
            discos_vms[nome_vm] = []

            for disco_remoto in maquina_virtual.get('Discos'):
                disco = VMDisco(
                    SCDiskBusType(disco_remoto.get('Tipo')),
                    disco_remoto.get('Arquivo'),
                    disco_remoto.get('TamanhoMB'),
                    SCDiskSizeType(disco_remoto.get('TamanhoTipo')),
                    disco_remoto.get('Caminho'))

                disco.set_parametros_extras_vmm(
                    disco_remoto.get('IDDrive'),
                    disco_remoto.get('IDDisco'),
                    disco_remoto.get('Bus'),
                    disco_remoto.get('Lun'),
                )

                discos_vms[nome_vm].append(disco)

        return discos_vms

    def __montar_inventario(self, servidor_acesso,
                            filtro_nome_vm=None, filtro_dados_completos=True):
        self.__inventario = Inventario(self.agrupamento, self.nuvem)

        vms_servidor = json.loads(
            self.__get_vms_servidor(servidor_acesso, filtro_nome_vm) or '{}')

        for maquina_virtual in vms_servidor:
            vms_rede = []
            for rede in maquina_virtual.get('Redes'):
                vm_rede = VMRede(rede.get('Nome'), rede.get('Principal'))
                vm_rede.ips = rede.get('IPS', '').split(' ')
                vms_rede.append(vm_rede)

            self.__inventario.vms[maquina_virtual.get('Nome')] = VM(
                maquina_virtual.get('Nome'),
                maquina_virtual.get('Descricao'),
                maquina_virtual.get('Imagem'),
                maquina_virtual.get('Regiao'),
                maquina_virtual.get('QtdeCpu'),
                maquina_virtual.get('QtdeRam'),
                vms_rede,
                maquina_virtual.get('ID'),
                maquina_virtual.get('VirtualizacaoAninhada'),
                maquina_virtual.get('MemoriaDinamica'),
                VMStatusEnum(maquina_virtual.get('Status')),
                maquina_virtual.get('NoRegiao'),
            )

        # Obtendo dados adicionais
        if filtro_dados_completos:
            # discos
            self.__inventario.set_discos_vms(
                self.__get_discos_adicionais(servidor_acesso))

            # regioes
            self.__inventario.set_regioes_disponiveis(
                ParserRemoto.__get_regioes_disponiveis(servidor_acesso))

    def get_inventario(self, servidor_acesso, filtro_nome_vm=None, filtro_dados_completos=True):
        if not self.__inventario:
            try:
                self.__montar_inventario(
                    servidor_acesso, filtro_nome_vm, filtro_dados_completos)
            # pylint: disable=broad-except
            except Exception as ex:
                return False, str(ex)

        return True, self.__inventario



"""
Representação de um disco adicional uma máquina virtual
"""
from vmm_manager.entidade.acao import Acao


class VMDisco:
    def __init__(self, tipo, arquivo, tamanho_mb, tamanho_tipo, caminho=None):
        self.tipo = tipo
        self.arquivo = arquivo
        self.tamanho_mb = tamanho_mb
        self.tamanho_tipo = tamanho_tipo
        self.caminho = caminho

        self.__id_drive = None
        self.__id_disco = None
        self.__bus = None
        self.__lun = None

    def set_parametros_extras_vmm(self, id_drive, id_disco, bus, lun):
        self.__id_drive = id_drive
        self.__id_disco = id_disco
        self.__bus = bus
        self.__lun = lun

    def get_id_disco(self):
        if not self.__id_disco:
            raise ValueError('Disco local não possui ID.')

        return self.__id_disco

    def get_id_drive(self):
        if not self.__id_drive:
            raise ValueError('Disco local não possui ID.')

        return self.__id_drive

    def get_tamanho_tipo_create(self):
        return self.tamanho_tipo.name

    def get_acao_criar_disco(self, nome_vm):
        return Acao('criar_disco_vm',
                    nome_vm=nome_vm,
                    tipo=self.tipo.value,
                    tamanho_mb=self.tamanho_mb,
                    tamanho_tipo=self.get_tamanho_tipo_create(),
                    arquivo=self.arquivo,
                    caminho=self.caminho)

    def get_acao_excluir_disco(self, id_vm):
        return Acao(Acao.ACAO_EXCLUIR_DISCO_VM,
                    id_vm=id_vm,
                    id_drive=self.get_id_drive())

    def get_acao_expandir_disco(self, id_vm, id_drive):
        return Acao('expandir_disco_vm',
                    id_vm=id_vm,
                    id_drive=id_drive,
                    tamanho_mb=self.tamanho_mb)

    def get_acao_mover_disco(self, id_vm, id_disco):
        return Acao('mover_disco_vm',
                    id_vm=id_vm,
                    id_disco=id_disco,
                    caminho=self.caminho)

    def get_acao_converter_disco(self, id_vm, id_drive):
        return Acao('converter_disco_vm',
                    id_vm=id_vm,
                    id_drive=id_drive,
                    tamanho_tipo=self.get_tamanho_tipo_create())

    def get_acoes_diferenca_disco(self, disco_remoto, id_vm, nome_vm):
        acoes = []

        # Alteração de tipo ou redução de disco exige a recriação
        if ((self.tipo != disco_remoto.tipo) or
                (self.tamanho_mb < disco_remoto.tamanho_mb)):
            acoes.append(disco_remoto.get_acao_excluir_disco(id_vm))
            acoes.append(self.get_acao_criar_disco(nome_vm))
        else:
            # Tipo de tamanho alterado: converter disco
            if self.tamanho_tipo != disco_remoto.tamanho_tipo:
                acoes.append(self.get_acao_converter_disco(
                    id_vm, disco_remoto.get_id_drive()))

            # Tamanho alterado: expansão
            if self.tamanho_mb > disco_remoto.tamanho_mb:
                acoes.append(self.get_acao_expandir_disco(
                    id_vm, disco_remoto.get_id_drive()))

            # Caminho alterado: mover disco
            if self.caminho and self.caminho != disco_remoto.caminho:
                acoes.append(self.get_acao_mover_disco(
                    id_vm, disco_remoto.get_id_disco()))

        return acoes

    def __hash__(self):
        return hash(self.arquivo)

    def __eq__(self, other):
        return isinstance(other, VMDisco) and (self.tipo == other.tipo
                                               and self.arquivo == other.arquivo
                                               and self.tamanho_mb == other.tamanho_mb
                                               and self.tamanho_tipo == other.tamanho_tipo
                                               and self.caminho == other.caminho)

    def __repr__(self):
        return f'''
                tipo: {self.tipo}
                arquivo: {self.arquivo}
                tamanho_mb: { self.tamanho_mb}
                tamanho_tipo: { self.tamanho_tipo}
                caminho: {self.caminho}
                id_drive: {self.__id_drive}
                id_disco: {self.__id_disco}
                bus: {self.__bus}
                lun: {self.__lun}
                '''

    def to_dict(self):
        return {
            'tipo': self.tipo.value,
            'arquivo': self.arquivo,
            'tamanho_mb': self.tamanho_mb,
            'tamanho_tipo': self.tamanho_tipo.value,
            'caminho': self.caminho,
            'id_drive': self.__id_drive,
            'id_disco': self.__id_disco,
            'bus': self.__bus,
            'lun': self.__lun
        }



"""
Representação de um nó de um cluster no system center virtual machine manager.
Para seguir a nomemclatura de nuvem, designado aqui como 'Region'
"""


# pylint: disable=too-few-public-methods
class SCRegion():
    REGIAO_PADRAO = 'default'

    def __init__(self, id_no, nome_no, grupo, cluster, letra_id=None):
        self.id_no = id_no
        self.nome_no = nome_no
        self.grupo = grupo
        self.cluster = cluster
        self.letra_id = letra_id

    def __str__(self):
        return f'''
            id_no: {self.id_no}
            nome_no: {self.nome_no}
            grupo: {self.grupo}
            cluster: {self.cluster}
            letra_id: {self.letra_id}
            '''



"""
Representação de um plano de execução
"""
import os
import uuid
import textwrap
from yamlable import yaml_info, YamlAble
import yaml
from vmm_manager.infra.comando import Comando
from vmm_manager.util.msgs import imprimir_acao_corrente, formatar_msg_erro
from vmm_manager.util.msgs import imprimir_erro, imprimir_ok
from vmm_manager.scvmm.scjob import SCJob
from vmm_manager.entidade.acao import Acao


@yaml_info(yaml_tag_ns='scvmm_manager')
class PlanoExecucao(YamlAble):
    ARQUIVO_PLANO_EXECUCAO = 'plan.yaml'
    ARQUIVO_LOG_ERROS = 'erros.log'

    @staticmethod
    def carregar_plano_execucao(arquivo_plano_execucao):
        try:
            with open(arquivo_plano_execucao, 'r', encoding='utf8') as arquivo:
                plano_execucao = yaml.safe_load(arquivo)
                return True, plano_execucao
        except (IOError, TypeError) as erro:
            return False, f"Erro ao carregar plano de execução '{arquivo_plano_execucao}'.\n{erro}"

    @staticmethod
    def excluir_arquivo():
        if os.path.exists(PlanoExecucao.ARQUIVO_PLANO_EXECUCAO):
            os.remove(PlanoExecucao.ARQUIVO_PLANO_EXECUCAO)

    @staticmethod
    def __excluir_arquivo_log_erros():
        if os.path.exists(PlanoExecucao.ARQUIVO_LOG_ERROS):
            os.remove(PlanoExecucao.ARQUIVO_LOG_ERROS)

    def __init__(self, agrupamento, nuvem):
        self.agrupamento = agrupamento
        self.nuvem = nuvem
        self.acoes = []
        self.__jobs_em_execucao = {}
        self.__guids_a_limpar = []
        self.__msgs_erros = ''

    def is_vazio(self):
        return not self.acoes

    def gerar_arquivo(self):
        try:
            conteudo = yaml.safe_dump(self, default_flow_style=False)
            with open(PlanoExecucao.ARQUIVO_PLANO_EXECUCAO, 'w', encoding='utf8') as arquivo_yaml:
                arquivo_yaml.write(conteudo)
        except IOError as erro:
            return False, f'Erro ao gerar arquivo {PlanoExecucao.ARQUIVO_PLANO_EXECUCAO}.\n{erro}'

        return True, conteudo

    def __executar_acoes(self, acoes, servidor_acesso, ocultar_progresso):
        if acoes:
            # zerando jobs em execução
            self.__jobs_em_execucao = {}

            for acao in acoes:
                print(f'{textwrap.shorten(acao.get_str_impressao_inline(), 100)} => ',
                      end='', flush=True)

                # Caso específico de criação de vms
                guid = uuid.uuid4()
                if acao.is_criacao_vm():
                    self.__guids_a_limpar.append(guid)

                acao.executar(self.agrupamento, self.nuvem,
                              servidor_acesso, guid)

                # Tratando resultado
                if acao.was_executada_com_sucesso():
                    guid = acao.get_resultado_execucao().get('Guid')
                    if guid:
                        # Job em execução
                        self.__jobs_em_execucao[guid] = SCJob(
                            guid, acao)
                        print(f'[Iniciado {guid}]')
                    else:
                        # Comando já finalizado
                        imprimir_ok(ocultar_progresso)
                else:
                    imprimir_erro(ocultar_progresso)
                    self.__logar_erros_acao(
                        acao, acao.get_resultado_execucao().get('Msgs'))

            # Monitorando jobs
            SCJob.monitorar_jobs(self.__jobs_em_execucao, servidor_acesso)
            self.__coletar_resultado_jobs()

            # Ações pós execução
            print()
            self.__executar_cmds_finalizacao(
                acoes, servidor_acesso, ocultar_progresso)

    def executar(self, servidor_acesso, ocultar_progresso):
        if not self.is_vazio():
            print('\nOperações executadas:')

            # acoes bloqueantes primeiro
            acoes_bloqueantes = [
                acao for acao in self.acoes if acao.is_bloqueante()]
            self.__executar_acoes(
                acoes_bloqueantes, servidor_acesso, ocultar_progresso)

            # demais ações, se não houve erro
            if not self.has_erro_execucao():
                acoes_nao_bloqueantes = [acao
                                         for acao in self.acoes if not acao.is_bloqueante()]
                self.__executar_acoes(
                    acoes_nao_bloqueantes, servidor_acesso, ocultar_progresso)

            # Ações de finalização
            self.__limpar_guids(servidor_acesso, ocultar_progresso)
            self.__processa_resultado_execucao()
            PlanoExecucao.excluir_arquivo()

    def __executar_cmds_finalizacao(self, acoes, servidor_acesso, ocultar_progresso):
        for acao in acoes:
            if acao.was_executada_com_sucesso() and acao.has_cmd_pos_execucao():
                cmd = acao.get_cmd_pos_execucao(
                    self.agrupamento, servidor_acesso)

                imprimir_acao_corrente(cmd.descricao, ocultar_progresso)

                status, resultado = cmd.executar(servidor_acesso)

                if status:
                    imprimir_ok(ocultar_progresso)
                else:
                    imprimir_erro(ocultar_progresso)
                    self.__logar_erros_comando(cmd.descricao, resultado)

    def __limpar_guids(self, servidor_acesso, ocultar_progresso):
        if self.__guids_a_limpar:
            imprimir_acao_corrente(
                'Limpando objetos temporários', ocultar_progresso)

            cmd = Comando('limpar_objs_criacao_vm',
                          servidor_vmm=servidor_acesso.servidor_vmm,
                          guids=self.__guids_a_limpar)
            status, retorno = cmd.executar(servidor_acesso)

            if status:
                imprimir_ok(ocultar_progresso)
            else:
                imprimir_erro(ocultar_progresso)
                self.__logar_erros_comando(
                    'limpar objetos temporários', retorno)

    def __processa_resultado_execucao(self):
        if self.has_erro_execucao():
            self.__gerar_arquivo_erros()
        else:
            PlanoExecucao.__excluir_arquivo_log_erros()

    def imprimir_resultado_execucao(self):
        if self.has_erro_execucao():
            print(formatar_msg_erro(
                '\nErro ao executar algumas operações. '
                f'Mais detalhes em {PlanoExecucao.ARQUIVO_LOG_ERROS}.'))
        else:
            print('\nSincronização executada com sucesso.')

    def has_erro_execucao(self):
        return self.__msgs_erros

    def __coletar_resultado_jobs(self):
        if self.__jobs_em_execucao:
            for job in self.__jobs_em_execucao.values():
                if job.is_finalizado_com_erro():
                    print(formatar_msg_erro(
                        f'Processo {job.id_vmm} finalizado com erro.'))
                    self.__logar_erros_acao(job.acao, job.resumo_erro)

    def __logar_erros_comando(self, comando, erro):
        self.__msgs_erros += f'Erro no comando "{comando}":\n{erro}\n\n'

    def __logar_erros_acao(self, acao, erro):
        self.__msgs_erros += f'Comando:\n{ acao.get_str_impressao_inline()}' \
            f'\n\nErro capturado:\n{erro}\n\n'

    def __gerar_arquivo_erros(self):
        try:
            with open(PlanoExecucao.ARQUIVO_LOG_ERROS, 'w', encoding='utf8') as arquivo_erros:
                arquivo_erros.write(self.__msgs_erros)
        except IOError as erro:
            print(
                f'Não foi possível gerar arquivo de erros ({PlanoExecucao.ARQUIVO_LOG_ERROS}): '
                f'{erro}')

    def imprimir_acoes(self):
        for acao in self.acoes:
            print(acao.get_str_impressao_inline())

    def __eq__(self, other):
        return isinstance(other, PlanoExecucao) and (self.agrupamento == other.agrupamento
                                                     and self.nuvem == other.nuvem
                                                     and self.acoes == other.acoes)

    def __str__(self):
        return f'''
            agrupamento: {self.agrupamento}
            nuvem: {self.nuvem}
            acoes: {','.join(str(acao) for acao in self.acoes)}
            '''

    def __to_yaml_dict__(self):
        return {'agrupamento': self.agrupamento,
                'nuvem': self.nuvem,
                'acoes': self.acoes}

    @ classmethod
    def __from_yaml_dict__(cls, dct, yaml_tag):
        plano_execucao = PlanoExecucao(dct['agrupamento'], dct['nuvem'])
        for acao_str in dct['acoes']:
            acao = Acao(acao_str.nome_comando)
            acao.args = acao_str.args['args']
            plano_execucao.acoes.append(acao)
        return plano_execucao



"""
Enums do VMM
"""
from enum import Enum


class VMStatusEnum(Enum):
    EM_EXECUCAO = 0
    DESLIGADA = 1
    FALHA_NA_ATUALIZACAO = 107


class SCJobStatusEnum(Enum):
    EM_EXECUCAO = 1
    FALHA = 3
    CANCELADO = 4
    SUCESSO = 5
    SUCESSO_COM_AVISO = 6


class SCDiskBusType(Enum):
    SCSI = 'SCSI'
    IDE = 'IDE'


class SCDiskSizeType(Enum):
    FIXED = 'FixedSize'
    DYNAMIC = 'DynamicallyExpanding'
